/* Ms.c - MandelSpawn popup widget */

/*  This file is part of MandelSpawn, a parallel Mandelbrot program for
    the X window system.

    Copyright (C) 1990 Andreas Gustafsson

    MandelSpawn is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License, version 1,
    as published by the Free Software Foundation.

    MandelSpawn is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License,
    version 1, along with this program; if not, write to the Free 
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* This widget creates a window showing a part of the Mandelbrot set */
/* It delegates the calculations to the slaves by calling the */
/* DispatchChunk() function in the Mama widget, and receives the */
/* results through the Draw() function. */

#include <stdio.h>
#include <math.h>

#include <X11/IntrinsicP.h>
#include <X11/Xos.h>
#include <X11/StringDefs.h>

#include "backward.h" /* X11R2 backward compatibility stuff */

#include "MsP.h" /* includes Ms.h which includes MsJob.h */
#include "Mama.h"

extern XtAppContext thisApp;
extern Display *myDisplay;
extern Screen *myScreen;

/* Misc. functions */
static void
    ClassInitialize(), Initialize(), Resize(), Realize(), Destroy(), 
    InvertBox(), EraseBox(), HideBox(), UnhideBox(),
    DoExpose();

/* These are used by menu choises in R4menu.c */
void ZoomIn(), WindowStats();

/* Action functions */
static void
    BeginBoxAction(), StretchBoxAction(), EndBoxAction(),
    ZoomAction(), CloseAction(), QuitAction(), ApplStatsAction(), 
    WindowStatsAction();

#ifdef MENU
void MsCreateMenu();
#endif
#ifdef LABEL
void MsCreateUnderflowLabel();
#endif

static Boolean SetValues();

/* Defaults */

static Dimension default_width = 400;	/* window width in pixels */
static Dimension default_height = 250;	/* window height in pixels */
static int default_iteration_limit = 0;	/* 0 means undefined */
static double default_center_x = (-0.5); /* x coordinate of window center */
static double default_center_y = 0.0;	/* y coordinate of window center */
static double default_range = 4.0; 	/* window range in x direction */
static Bool default_center_box = True;	/* do we center the rubberband box? */
static Bool default_julia = False;	/* do we show the Julia set? */
static double default_c_x = 0.0; 	/* c parameter for Julia, real */
static double default_c_y = 0.0;	/* c parameter for Julia, im */
static double default_julia_range = 4.0; /*  */
static double default_julia_center_x = 0.0; /*  */
static double default_julia_center_y = 0.0;	/*  */
static char default_cursor[] = "top_left_arrow"; 
static Dimension default_chunk_width = 32;
static Dimension default_chunk_height = 32;

extern char msDefaultTranslations[];

#ifndef MENU
/* Some of these bindings are less than obvious.  Don't complain, */
/* get X11R4 and use the popup menu in R4menu.c. */
char msDefaultTranslations[] = 
    "<Btn1Down>:	BeginBox()		\n\
     <Btn1Motion>:	StretchBox()		\n\
     <Btn1Up>:		EndBox()		\n\
     Shift<Btn2Down>:	Zoom(popup, nojulia)	\n\
     <Btn2Down>:	Zoom(nopop, nojulia)	\n\
     Shift<Btn3Down>:	Quit()			\n\
     <Key>q:		Quit()			\n\
     <Key>z:		Zoom(popup, nojulia)	\n\
     <Key>i:		Zoom(nopop, nojulia)	\n\
     <Key>s:		ApplStats()		\n\
     <Key>w:		WindowStats()		\n\
     <Key>j:		Zoom(popup, julia)	\n\
     <Btn3Down>:	Close()";
#endif

static XtActionsRec actionsList[] =
{
    { "BeginBox",	BeginBoxAction		},
    { "EndBox",		EndBoxAction		},
    { "StretchBox",	StretchBoxAction	},
    { "Zoom",		ZoomAction		},
    { "WindowStats",	WindowStatsAction	},
    { "ApplStats",	ApplStatsAction		},
    { "Close",		CloseAction		},
    { "Quit",		QuitAction		}
};

static XtResource resources[] = 
{   /* Core resources */
    { XtNwidth, XtCWidth, XtRDimension, sizeof(Dimension),
	XtOffset(Widget, core.width), XtRDimension,
	(caddr_t) &default_width },
    { XtNheight, XtCHeight, XtRDimension, sizeof(Dimension),
	XtOffset(Widget, core.height), XtRDimension,
	(caddr_t) &default_height },
    /* Noncore resources */
    { XtNiteration_limit, XtCValue, XtRInt, sizeof(int),
	XtOffset(MsWidget, ms.job.iteration_limit), XtRInt,
	(caddr_t) &default_iteration_limit },
    { XtNCenterX, XtCValue, XtRDouble, sizeof(double),
	XtOffset(MsWidget, ms.center_x), XtRDouble,
	(caddr_t) &default_center_x },
    { XtNCenterY, XtCValue, XtRDouble, sizeof(double),
	XtOffset(MsWidget, ms.center_y), XtRDouble,
	(caddr_t) &default_center_y },
    { XtNRange, XtCValue, XtRDouble, sizeof(double),
	XtOffset(MsWidget, ms.xrange), XtRDouble,
	(caddr_t) &default_range },
    { XtNCenterBox, XtCValue, XtRBool, sizeof(Bool),
	XtOffset(MsWidget, ms.center_box), XtRBool,
	(caddr_t) &default_center_box },
    { XtNCursor, XtCValue, XtRCursor, sizeof(Cursor),
	XtOffset(MsWidget, ms.my_cursor), XtRString,
	(caddr_t) default_cursor },
    { XtNMama, XtCValue, XtRPointer, sizeof(MamaWidget),
	XtOffset(MsWidget, ms.mama), XtRPointer,
	(caddr_t) 0 } ,
    { XtNJulia, XtCValue, XtRBool, sizeof(Bool),
	XtOffset(MsWidget, ms.julia), XtRBool,
	(caddr_t) &default_julia },
    { XtNCX, XtCValue, XtRDouble, sizeof(double),
	XtOffset(MsWidget, ms.c_x), XtRDouble,
	(caddr_t) &default_c_x },
    { XtNCY, XtCValue, XtRDouble, sizeof(double),
	XtOffset(MsWidget, ms.c_y), XtRDouble,
	(caddr_t) &default_c_y },
    { XtNChunkWidth, XtCValue, XtRDimension, sizeof(Dimension),
	XtOffset(MsWidget, ms.chunk_width), XtRDimension,
	(caddr_t) &default_chunk_width },
    { XtNChunkHeight, XtCValue, XtRDimension, sizeof(Dimension),
	XtOffset(MsWidget, ms.chunk_height), XtRDimension,
	(caddr_t) &default_chunk_height }
};

MsClassRec msClassRec = 
{   /* core fields */
    { 
    /* superclass		*/	(WidgetClass) &compositeClassRec,
    /* class_name		*/	"MandelSpawn",
    /* widget_size		*/	sizeof(MsRec),
    /* class_initialize		*/      NULL,
    /* class_part_initialize	*/	NULL,
    /* class_inited		*/	FALSE,
    /* initialize		*/	Initialize,
    /* initialize_hook		*/	NULL,
    /* realize			*/	Realize,
    /* actions			*/	actionsList,
    /* num_actions		*/	XtNumber(actionsList),
    /* resources		*/	resources,
    /* resource_count		*/	XtNumber(resources),
    /* xrm_class		*/	NULL,
    /* compress_motion		*/	TRUE,
    /* compress_exposure	*/	FALSE,
    /* compress_enterleave	*/	TRUE,
    /* visible_interest		*/	FALSE,
    /* destroy			*/	Destroy,
    /* resize			*/	Resize,
    /* expose			*/	DoExpose,
    /* set_values		*/	SetValues,
    /* set_values_hook		*/	NULL,
    /* set_values_almost	*/	XtInheritSetValuesAlmost,
    /* get_values_hook		*/	NULL,
    /* accept_focus		*/	NULL,
    /* version			*/	XtVersion,
    /* callback_private		*/	NULL,
    /* tm_table			*/	msDefaultTranslations,
    /* query_geometry		*/	NULL
    },
    /* CompositeClassPart fields */
    { XtInheritGeometryManager,
      XtInheritChangeManaged,
      XtInheritInsertChild,
      XtInheritDeleteChild
    },
    /* msClassPart fields */
    {
    /* dummy			*/	0
    }
};

WidgetClass msWidgetClass = (WidgetClass) &msClassRec;


/* Initialize the widget */

static void
Initialize(request, new)
     MsWidget   request;
     MsWidget   new;
{ XGCValues gcv;
  int depth = DefaultDepthOfScreen(XtScreen(new));
  /* set up a GC for the rubberband box */
  gcv.function=GXxor;
  
  /* "gcv.foreground=~0l" messes up the black-and-white mode of the */
  /* X11R2 server for the Sony NWS-1510 4-plane display quite badly; */
  /* I'd call that a server bug.  The code below works but assumes */
  /* 2's complement. */
  if(depth==32)
    gcv.foreground=0xffffffff; /* 1<<32 is undefined, says K&R */
  else
    gcv.foreground= (1 << depth) - 1;
  new->ms.box_gc=XtGetGC(new, GCFunction|GCForeground, &gcv);
  new->ms.mi_count = 0L; /* done no iterations yet */
  new->ms.box_exists=False;
  new->ms.box_hidden=0;
  new->ms.linebuffer=(XImage *) NULL; /* no scanline buffer so far */
  new->ms.chunks_out=0;
  new->ms.configuration=0;
#ifdef LABEL
  new->ms.underflow=0;
  new->ms.underflow_label=NULL;
#endif
}


/* Calculate the various parameters that go into ms.job.  This includes */
/* host to network real number format conversion. */

static void CalculateJobParameters(w, j)
     MsWidget w; struct static_job_info *j;
{ j->julia=w->ms.julia;
  if(j->julia) /* Julia */
  { j->z0.re=double_to_net(w->ms.center_x - w->ms.xrange / 2.0);
    j->z0.im=double_to_net(w->ms.center_y - w->ms.yrange / 2.0);
    j->corner.re=double_to_net(w->ms.c_x);
    j->corner.im=double_to_net(w->ms.c_y);
  }
  else /* Mandelbrot */
  { j->corner.re=double_to_net(w->ms.center_x - w->ms.xrange / 2.0);
    j->corner.im=double_to_net(w->ms.center_y - w->ms.yrange / 2.0);
    j->z0.re=double_to_net(0.0);
    j->z0.im=double_to_net(0.0);
  }
  j->delta.re=double_to_net(w->ms.xrange/w->core.width);
  j->delta.im=double_to_net(w->ms.yrange/w->core.height);
}


/* Initialize, or re-initialize widget fields that may need to be */
/* changed when the widget is resized, zoomed or otherwise modified */

static void MsPrecalculate(w) MsWidget w;
{ int underflow;
  w->ms.yrange=w->ms.xrange *
    ((double) w->core.height / (double) w->core.width);

  CalculateJobParameters(w, &w->ms.job);

#ifdef LABEL
  underflow=(w->ms.job.delta.re == 0 || w->ms.job.delta.im == 0);

  if(underflow && !w->ms.underflow)
  { MsCreateUnderflowLabel(w);
  }
  if(!underflow && w->ms.underflow)
  { XtDestroyWidget(w->ms.underflow_label);
  }
  w->ms.underflow = underflow;
#endif

  /* if it had a scanline  buffer already, destroy it */
  if(w->ms.linebuffer)
    XDestroyImage(w->ms.linebuffer);

  /* create the scanline buffer */
  w->ms.linebuffer=
    XCreateImage(XtDisplay(w), DefaultVisualOfScreen(XtScreen(w)), 
		 DefaultDepthOfScreen(XtScreen(w)), ZPixmap, 0, (char *) NULL, 
		 w->ms.chunk_width, w->ms.chunk_height, 32,
		 0 /* zero means let XCreateImage determine bytes/line */
		);
  w->ms.linebuffer_size=(unsigned) w->ms.linebuffer->bytes_per_line * 
    w->ms.chunk_height;
  w->ms.linebuffer->data=(char *) XtMalloc(w->ms.linebuffer_size);

  if(w->ms.job.iteration_limit == 0) /* not set yet? */
  { /* iterate as far as possible by default */
    w->ms.job.iteration_limit = MaxIterations(w->ms.mama);
  }
  else
    if(w->ms.job.iteration_limit > MaxIterations(w->ms.mama))
    { w->ms.job.iteration_limit=MaxIterations(w->ms.mama);
      XtAppWarning(thisApp,
		   "Iteration limit truncated");
    }

  w->ms.bytes_per_count=(w->ms.job.iteration_limit>256 ? 2 : 1);
  w->ms.pixels_per_chunk=MaxMessageSize() / w->ms.bytes_per_count;

  /* determine the kind of draw routine to use depending on */
  /* display characteristics and iteration count size */

  if(w->ms.linebuffer->bits_per_pixel == 8)
    w->ms.type=DISP_8plane;
  else if(w->ms.linebuffer->bits_per_pixel == 1  &&
	    w->ms.linebuffer->bitmap_unit==32 &&
	    w->ms.linebuffer->bitmap_bit_order != LSBFirst)
      w->ms.type=DISP_1plane32msb; 
  else
  { w->ms.type=DISP_generic;
    XtAppWarning(thisApp,
		 "No special case support for this display type;\n\
drawing a pixel at a time, this will be very slow.");
  }

  w->ms.type |= (w->ms.bytes_per_count==2) ? ITER_WORD : ITER_BYTE;

  /* We now have a new configuration; give it a unique number */
  w->ms.configuration++;
}


/* Calculate the coordinates for the rubberband box given the initial */
/* and current mouse position (box_origin, box_corner). Use the box_origin */
/* as either the center or the oppsite corner depending on the center_box */
/* flag */

struct box UnflipBox(w)
     MsWidget w;
{ struct box r;
  if(w->ms.center_box)
  { int halfwidth=ABS(w->ms.box_corner.x - w->ms.box_origin.x);
    int halfheight=ABS(w->ms.box_corner.y - w->ms.box_origin.y);
    r.x0=w->ms.box_origin.x-halfwidth;
    r.x1=w->ms.box_origin.x+halfwidth;
    r.y0=w->ms.box_origin.y-halfheight;
    r.y1=w->ms.box_origin.y+halfheight;
  }
  else
  { r.x0=MIN(w->ms.box_origin.x, w->ms.box_corner.x);
    r.x1=MAX(w->ms.box_origin.x, w->ms.box_corner.x);
    r.y0=MIN(w->ms.box_origin.y, w->ms.box_corner.y);
    r.y1=MAX(w->ms.box_origin.y, w->ms.box_corner.y);
  }
  return(r);
}


/* Zoom into the area selected using the box.  If "pop", pop up a new */
/* window.  If "julia", show the Julia set corresponding to the point */
/* at the center of the box. */

void ZoomIn(w, pop, julia)
       MsWidget w;
       int pop;
       int julia;
{ struct box b;
  double scale;
  double new_xrange, new_center_x, new_center_y;
  Arg arglist[12];
  int num_args;
  Arg shell_arglist[12];
  int num_shell_args;

  if(!w->ms.box_exists)
    return;

  b=UnflipBox(w);  

  scale=(double) (b.x1-b.x0) / (double) (w->core.width);
  new_xrange=w->ms.xrange*scale;
  new_center_x=w->ms.center_x+
    ((((double) b.x0+b.x1)/2)-(w->core.width/2))/
      (double)(w->core.width)*(w->ms.xrange);
  new_center_y=w->ms.center_y+
    ((((double) b.y0+b.y1)/2)-(w->core.height/2))/
      (double)(w->core.height)*(w->ms.yrange);

  /* build arguments for the changed resources in the zoomed widget */
  num_args=0;

  /* Transition from Mandelbrot to Julia set */
  if(julia && !(w->ms.julia))
  { XtSetArg(arglist[num_args], XtNJulia, True); num_args++;
    /* Set the C parameter (chooses a Julia set out of infinitely many) */
    XtSetArg(arglist[num_args], XtNCX, &new_center_x); num_args++;
    XtSetArg(arglist[num_args], XtNCY, &new_center_y); num_args++;
    /* This is an initial Julia picture, so show the whole Julia set */
    XtSetArg(arglist[num_args], XtNRange, &default_julia_range); num_args++;
    XtSetArg(arglist[num_args], XtNCenterX,
	     &default_julia_center_x); num_args++;
    XtSetArg(arglist[num_args], XtNCenterY,
	     &default_julia_center_y); num_args++;
  }
  else /* No change in M/J mode */
  { XtSetArg(arglist[num_args], XtNJulia, w->ms.julia); num_args++;
    XtSetArg(arglist[num_args], XtNCenterX, &new_center_x); num_args++;
    XtSetArg(arglist[num_args], XtNCenterY, &new_center_y); num_args++;
    XtSetArg(arglist[num_args], XtNRange, &new_xrange); num_args++;
    /* these two parameters are redundant in the Mandelbrot mode */
    XtSetArg(arglist[num_args], XtNCX, &w->ms.c_x); num_args++;
    XtSetArg(arglist[num_args], XtNCY, &w->ms.c_y); num_args++;
  }

  XtSetArg(arglist[num_args], XtNMama, w->ms.mama); num_args++;

  if(pop)
  { num_shell_args = 0;
    /* make the new window as big as this one initially */
    XtSetArg(shell_arglist[num_shell_args], XtNheight, w->core.height);
    num_shell_args++;
    XtSetArg(shell_arglist[num_shell_args], XtNwidth, w->core.width);
    num_shell_args++;
    /* create a new pop-up window for the zoomed area */
    PopupAnother(w->ms.mama, shell_arglist, num_shell_args, arglist, num_args);
  }
  else 
  { /* zoom using the old window: just set the changed resources */
    EraseBox(w);
    XtSetValues(w, arglist, num_args);
  }
}


/* Action routine interface to ZoomIn */

static void ZoomAction(w, event, params, nparams)
     MsWidget w;
     XEvent *event;
     String *params;
     Cardinal *nparams;
{ 
  if(*nparams != 2)
    return;
  ZoomIn(w, params[0][0] == 'p', params[1][0] == 'j');
}


/* Take a rectangle to be updated, build a "job packet" for */
/* calculating it and hand it over to the Mama widget */

void DispatchMandelRect(w, rect)
     MsWidget w;
     XRectangle rect;
{ ms_job j;
  ms_client_info info;
  info.configuration=w->ms.configuration;
  info.s=rect; /* structure assignment */
  w->ms.chunks_out++; /* one more to wait for */
  /* build a job structure in network byte order */
  j.j.julia=htons(w->ms.job.julia);
  j.j.corner.re=htonl(w->ms.job.corner.re);
  j.j.corner.im=htonl(w->ms.job.corner.im);
  j.j.z0.re=htonl(w->ms.job.z0.re);
  j.j.z0.im=htonl(w->ms.job.z0.im);
  j.j.delta.re=htonl(w->ms.job.delta.re);
  j.j.delta.im=htonl(w->ms.job.delta.im);
  j.j.iteration_limit=htonl(w->ms.job.iteration_limit);
  j.s.x=htons(rect.x);
  j.s.width=htons(rect.width);
  j.s.y=htons(rect.y);
  j.s.height=htons(rect.height);
  /* ..and put it on the work queue */
  DispatchChunk(w->ms.mama, w,
		(caddr_t) &info, sizeof(info),
		(caddr_t) &j, sizeof(j));
}


/* Handle widget exposure */
/* Can't call this "Expose" because that is #defined to 12 in X.h! */

/*ARGSUSED*/
static void DoExpose (w, e, r)
     MsWidget    w;
     XEvent *e;
     Region r;
{
  XRectangle ChunkRect;
  int i;

  unsigned int right_edge=e->xexpose.x+e->xexpose.width; 
  unsigned int bottom_edge=e->xexpose.y+e->xexpose.height; 
  unsigned int x, y;

#ifdef LABEL
  /* It's no use redrawing if underflow has occured */
  if(w->ms.underflow)
    return;
#endif

  /* remove the box for the duration of the update */
  if(w->ms.chunks_out==0) HideBox(w);
  
  for(y=e->xexpose.y; y < bottom_edge; y+=w->ms.chunk_height) 
    for(x=e->xexpose.x; x < right_edge; x+=w->ms.chunk_width) 
    { ChunkRect.x = x;
      ChunkRect.y = y;
      ChunkRect.width = MIN(w->ms.chunk_width, right_edge-x);
      ChunkRect.height = MIN(w->ms.chunk_height, bottom_edge-y);
      DispatchMandelRect(w, ChunkRect);
    }

  WhipAllSlaves(w->ms.mama);
}


/* Draw (or undraw) the rubberband box */

static void InvertBox(w)
     MsWidget w;
{ struct box b;
  /* draw only if no update is in progress */
  if(!w->ms.box_hidden)
  { b=UnflipBox(w);
    XDrawRectangle(XtDisplay(w), XtWindow(w), w->ms.box_gc,
		   b.x0, b.y0, b.x1-b.x0, b.y1-b.y0);
    w->ms.box_exists = !(w->ms.box_exists);
  }
}


/* Erase the rubberband box if it exists */

static void EraseBox(w)
     MsWidget w;
{ if(w->ms.box_exists) InvertBox(w);
}



/* Create the rubberband box */

static void BeginBoxAction(w, event, params, nparams)
     MsWidget w;
     XEvent *event;
     String *params;
     Cardinal *nparams;
{ EraseBox(w);
  w->ms.box_origin.x=w->ms.box_corner.x=((XButtonEvent *) event)->x;
  w->ms.box_origin.y=w->ms.box_corner.y=((XButtonEvent *) event)->y;
  InvertBox(w);
}


/* Stretch the rubberband box */

static void StretchBoxAction(w, event, params, nparams)
     MsWidget w;
     XEvent *event;
     String *params;
     Cardinal *nparams;
{ EraseBox(w);
  w->ms.box_corner.x=((XButtonEvent *) event)->x;
  w->ms.box_corner.y=((XButtonEvent *) event)->y;
  InvertBox(w);
}


/* Stop stretching the rubberband box (currently a no-op) */

static void EndBoxAction(w, event, params, nparams)
     MsWidget w;
     XEvent *event;
     String *params;
     Cardinal *nparams;
{
}


/* Hide the box when beginning an asynchronous update */

static void HideBox(w)
     MsWidget w;
{ if(w->ms.box_hidden==0 && w->ms.box_exists)
    InvertBox(w);
  w->ms.box_hidden++;
}


/* Unhide the box when an asynchronous update is done */

static void UnhideBox(w)
     MsWidget w;
{ w->ms.box_hidden--;
  if(w->ms.box_hidden==0 && w->ms.box_exists) 
    InvertBox(w);
}


/* Realize the widget */

static void
Realize (w, valueMask, attrs)
     MsWidget w;
     XtValueMask *valueMask;
     XSetWindowAttributes *attrs;
{ /* setting the colormap here is probably redundant */
  attrs->backing_store=Always;
  attrs->save_under=False;
  attrs->bit_gravity=ForgetGravity;
  attrs->cursor=w->ms.my_cursor;
  attrs->colormap=MamaColormap(w->ms.mama);
  XtCreateWindow (w, InputOutput, (Visual *) CopyFromParent,
    *valueMask | CWBackingStore | CWSaveUnder| CWBitGravity | CWCursor |
		  CWColormap, attrs);
  MsPrecalculate(w);
#ifdef MENU
  MsCreateMenu(w, !w->ms.julia);
#endif
}


/* Destroy the widget */

static void
Destroy (w)
     MsWidget w;
{ XtReleaseGC(w, w->ms.box_gc);
  XtFree((char *) w->ms.linebuffer);
}


/* Update widget resources */

static Boolean SetValues(current, request, new)
     MsWidget current, request, new;
{ MsPrecalculate(new); /* do the dirty work */
  return(True); /* widget must be redisplayed */
}
 

/* Resize the widget */

static void
Resize(w)
     MsWidget w;
{ MsPrecalculate(w); /* do the dirty work */
}


/* Print the view coordinates */

void WindowStats(w)
     MsWidget w;
{ (void) printf("current picture area: x = %f .. %f, y = %f .. %f\n",
		w->ms.center_x - w->ms.xrange / 2,
		w->ms.center_x + w->ms.xrange / 2,
		w->ms.center_y - w->ms.yrange / 2,
		w->ms.center_y + w->ms.yrange / 2
	       );
  (void) printf("return here with: ms ");
  if(w->ms.julia)
    (void) printf("-julia -cx %f -cy %f ", w->ms.c_x, w->ms.c_y);
  (void) printf("-x %f -y %f -range %f\n",
	       w->ms.center_x, w->ms.center_y, w->ms.xrange);
}

/* Action routine for the above */

static void WindowStatsAction(w, event, params, nparams)
     MsWidget w;
     XEvent *event;
     String *params;
     Cardinal *nparams;
{ WindowStats(w);
}


/* Print slave performance statistics */

static void ApplStatsAction(w, event, params, nparams)
     MsWidget w;
     XEvent *event;
     String *params;
     Cardinal *nparams;
{ SlaveStatistics(w->ms.mama);
}


/* This function is called from the Mama widget when a scan line has */
/* been completed by a slave */

void Draw(w, client_data, data, pixels)
     MsWidget w;
     caddr_t client_data;
     caddr_t data; 
     unsigned long *pixels;
{ register int i, j, xx;
  unsigned int x, y, width, height;

  ms_client_info *the_info = (ms_client_info *) client_data;

  /* Ignore the reply if the widget has changed shape or something */
  if(the_info->configuration == w->ms.configuration)
  { x=the_info->s.x; 
    y=the_info->s.y; 
    width=the_info->s.width;
    height=the_info->s.height;

    /* 16-bit iteration counts from the slave need to be */
    /* converted from network byte order.  It is faster */
    /* to do it all at once here than to call ntohs() */
    /* in the tight loops below.  Any decent compiler will */
    /* optimize away the whole "if" statement if the CPU */
    /* is big-endian; if it is little-endian then one */
    /* redundant ntohs() call will remain in the "if" condition. */

    if(ntohs(1) != 1)
    { if(w->ms.bytes_per_count == sizeof(unsigned short))
      {	unsigned short *end;
	register unsigned char *p;
	end = (unsigned short *) data + (width * height);
	for(p = (unsigned char *) data; p < (unsigned char *) end; p += 2)
	{ unsigned char tmp;
	  tmp = *p;
	  *p = *(p+1);
	  *(p+1) = tmp;
	}
      }
    }

    /* sorry about the combinatorial explosion, but this part easily */
    /* becomes a bottleneck if all the tests are done inside the loop */
    switch(w->ms.type)
    { 
    case ITER_BYTE|DISP_8plane:
      /* Quick-and-dirty drawing of 8-bit values on 8-plane displays */
      { unsigned char *datap = (unsigned char *) data;
	for(j=0; j<height; j++)
	{ unsigned char *bufp= (unsigned char *)
	    w->ms.linebuffer->data + (w->ms.linebuffer->bytes_per_line * j);
	  for(i=0; i<width; i++)
	  { *bufp++ = pixels[*datap++];
	  }
	}
      }
      break;
    case ITER_BYTE|DISP_1plane32msb:
      { unsigned char *datap = (unsigned char *) data;
	for(j=0; j<height; j++)
	{ unsigned long *bufp=
	    (unsigned long *)((unsigned char *)
	     w->ms.linebuffer->data + (w->ms.linebuffer->bytes_per_line * j));
	  for(i=0; i<width; ) /* for each word */
	  { unsigned long pixword = 0L;
	    unsigned long mask = 0x80000000L;
	    unsigned end=MIN(width, i+32);
	    for(; i<end; i++) /* for each bit in the word */
	    { if(pixels[*datap++])
		pixword |= mask;
	      mask >>= 1;
	    }
	    *bufp++ = htonl(pixword);
	  }
	}
      }
      break;
    case ITER_BYTE|DISP_generic:
      /* Slow but portable drawing of 8-bit values on 8-plane displays */
      { unsigned char *datap = (unsigned char *) data;
	for(j=0; j<height; j++)
	{ for(i=0; i<width; i++)
	    XPutPixel(w->ms.linebuffer, i, j, pixels[*datap++]);
	}
      }
      break;
    case ITER_WORD|DISP_8plane:
      /* Quick-and-dirty drawing of 16-bit values on 8-plane displays */
      { unsigned short *datap = (unsigned short *) data;
	for(j=0; j<height; j++)
	{ unsigned char *bufp= (unsigned char *)
	    w->ms.linebuffer->data + (w->ms.linebuffer->bytes_per_line * j);
	  for(i=0; i<width; i++)
	  { *bufp++ = pixels[*datap++];
	  }
	}
      }
      break;
    case ITER_WORD|DISP_1plane32msb:
      { unsigned short *datap = (unsigned short *) data;
	for(j=0; j<height; j++)
	{ unsigned long *bufp=
	    (unsigned long *)((unsigned char *)
	     w->ms.linebuffer->data + (w->ms.linebuffer->bytes_per_line * j));
	  for(i=0; i<width; ) /* for each word */
	  { unsigned long pixword = 0L;
	    unsigned long mask = 0x80000000L;
	    unsigned end=MIN(width, i+32);
	    for(; i<end; i++) /* for each bit in the word */
	    { if(pixels[*datap++])
		pixword |= mask;
	      mask >>= 1;
	    }
	    *bufp++ = htonl(pixword);
	  }
	}
      }
      break;
    case ITER_WORD|DISP_generic:
      /* Slow but portable drawing of 16-bit values on 8-plane displays */
      { unsigned short *datap = (unsigned short *) data;
	for(j=0; j<height; j++)
	{ for(i=0; i<width; i++)
	    XPutPixel(w->ms.linebuffer, i, j, pixels[*datap++]);
	}
      }
      break;
    default:
      XtAppError(thisApp, "no drawing routine defined for this display type");
    }
    /* go do it! */
    XPutImage(XtDisplay(w), XtWindow(w), DefaultGCOfScreen(XtScreen(w)),
	      w->ms.linebuffer, 
	      0, 0, /* position in the buffer */
	      x, y, /* position on the screen */
	      width, height);
  }
  w->ms.chunks_out--; /* one less to go */
  if(w->ms.chunks_out==0)
    UnhideBox(w);
}


/* Destroy this window (only) */

static void
CloseAction(w, event, params, nparams)
     MsWidget w;
     XEvent *event;
     String *params;
     Cardinal *nparams;
{ XtDestroyWidget(XtParent(w)); /* destroy the shell widget */
}


/* Ask Mama to shut down */

static void 
QuitAction(w, event, params, nparams)
     MsWidget w;
     XEvent *event;
     String *params;
     Cardinal *nparams;
{ Shutdown(w->ms.mama);
}
